Imports System.Drawing
Imports System.Drawing.Imaging
Imports System.Runtime.InteropServices

''' <summary>
''' verschafft sich Zugriff auf die Pixeldaten von Images, in Form von 2-dim Byte-Arrays
''' </summary>
Public Class PixelHelper
    'Der bliche Zugang zu Pixeldaten ber Bitmap.LockBits() impliziert 4 Kopiervorgnge der
    '! gesamten Bitmap: Lock/Unlock, und dann jeweils Marshall.Copy() in ein managed Array.
    'Die Bitmap-Erstellung unter Verwendung gepinnter Arrays reduziert die Kopiervorgnge auf
    '! zwei.

    Private _ReadPixels As Byte(,)
    Private _WritePixels As Byte(,)
    Private _ResultData As New BitmapData


    Private Shared _GrayMatrix As New Imaging.ColorMatrix(New Single()() _
                   {New Single() {0.299, 0.299, 0.299, 0, 0}, _
                    New Single() {0.587, 0.587, 0.587, 0, 0}, _
                    New Single() {0.114, 0.114, 0.114, 0, 0}, _
                    New Single() {0, 0, 0, 1, 0}, _
                    New Single() {0, 0, 0, 0, 1}})

    Private Shared _GrayAttr As New Imaging.ImageAttributes()

    Shared Sub New()
        _GrayAttr.SetColorMatrix(_GrayMatrix)
    End Sub

    Public Sub New( _
          ByVal Img As Image, _
          ByVal Format As Imaging.PixelFormat, _
          Optional ByVal ConvertToGray As Boolean = False)
        With _ResultData
            .Width = Img.Width
            .Height = Img.Height
            _ReadPixels = CreateWritePixels(Format)
            Dim GCH As GCHandle = GCHandle.Alloc(_ReadPixels, GCHandleType.Pinned)
            Using _
                  Bmp As New Bitmap(.Width, .Height, .Stride, Format, GCH.AddrOfPinnedObject()), _
                  G As Graphics = Graphics.FromImage(Bmp)
                Dim Rct As New Rectangle(0, 0, .Width, .Height)
                If ConvertToGray Then
                    G.DrawImage(Img, Rct, 0, 0, .Width, .Height, GraphicsUnit.Pixel, _GrayAttr)
                Else
                    G.DrawImageUnscaledAndClipped(Img, Rct)
                End If
            End Using
            GCH.Free()
        End With
    End Sub

    Public ReadOnly Property ReadPixels() As Byte(,)
        Get
            Return _ReadPixels
        End Get
    End Property

    Public Function CreateWritePixels(ByVal Format As Imaging.PixelFormat) As Byte(,)
        With _ResultData
            .PixelFormat = Format
            .Stride = GetStride(.Width, Format)
            ReDim _WritePixels(.Height - 1, .Stride - 1)
        End With
        Return _WritePixels
    End Function

    Public Function GetResultBitmap() As Bitmap
        With _ResultData
            Dim GCH As GCHandle = GCHandle.Alloc(_WritePixels, GCHandleType.Pinned)
            .Scan0 = GCH.AddrOfPinnedObject
            Dim Bmp As New Bitmap(.Width, .Height, .PixelFormat)
            Bmp.LockBits( _
               New Rectangle(0, 0, .Width, .Height), _
               ImageLockMode.WriteOnly Or ImageLockMode.UserInputBuffer, .PixelFormat, _ResultData)
            'UnlockBits() kopiert _ResultData in die Bitmap
            Bmp.UnlockBits(_ResultData)
            GCH.Free()
            Return Bmp
        End With
    End Function

    ''' <summary>konvertiert ein Image ins angegebene Pixelformat</summary>
    Public Shared Function Convert(ByVal Img As Image, ByVal Format As Imaging.PixelFormat) As Bitmap
        With Img
            Dim Bmp As New Bitmap(.Width, .Height, Format)
            Using G As Graphics = Graphics.FromImage(Bmp)
                G.DrawImageUnscaledAndClipped(Img, New Rectangle(0, 0, .Width, .Height))
            End Using
            Return Bmp
        End With
    End Function

    Public Shared Function GetStride( _
            ByVal Width As Integer, ByVal PixelFormat As Imaging.PixelFormat) As Integer
        'Aus Namen der Imaging.PixelFormat-Enumeration wie "Format16bppGrayScale" kann man die
        '! PixelSize (16bpp) direkt ausparsen
        Dim Name As String = EnumGetName(PixelFormat)
        Dim Indx As Integer = Name.IndexOf("b"c)
        If Indx < 0 Then Throw New Exception(String.Concat( _
           "Kann aus dem PixelFormat '", PixelFormat, "' keine PixelSize (bpp) ermitteln.", _
           "Whlen Sie ein System.Drawing.Imaging.PixelFormat, dass diese Angabe im Namen enthlt"))
        If Indx > 0 Then
            Dim _PixelSize As Integer = Integer.Parse(Name.Substring(6, Indx - 6))
            'Format-unabhngig vergrert eine Bildzeile (Stride) sich immer in 4-Byte-Schritten
            Return ((Width * _PixelSize + 31) And Not 31) \ 8
        End If
    End Function

End Class