Imports System.Drawing.Imaging
Imports System.IO
Imports ConVolutionFilter.FilterDataSet
Imports PictureData = System.Collections.Generic.KeyValuePair(Of String, System.Drawing.Bitmap)

Public Class frmConVolutionFilter
    Private _DataFile As String = "DataFile.xml"
    Private _Pictures As New List(Of PictureData)

    Private Sub frmConVolutionFilter_Load(ByVal sender As Object, ByVal e As EventArgs) Handles MyBase.Load
        AddHandler FilterDataSet.Filter.TableNewRow, AddressOf FilterTable_TableNewRow
        Dim PicDir As String = Path.GetFullPath("Pictures")
        Dim PicPathes As String() = Directory.GetFiles(PicDir)
        'Alle Bilder in KeyValuePairs laden, und per Databinding an cmbSourcePic binden
        For Each sPath As String In PicPathes
            Try
                Dim S As String = Path.GetFileName(sPath)
                _Pictures.Add(New PictureData(S, New Bitmap(sPath)))
            Catch ex As Exception
                MsgBox(String.Concat("Das Bild", Lf, sPath, Lf, "konnte nicht erstellt werden", Lf, _
                   "Vermutlich ist es keine Bilddatei, und sollte aus dem Bilderordner entfernt werden"))
                Diagnostics.Process.Start(PicDir)
            End Try
        Next
        Me.cmbSourcePic.DisplayMember = "Key"
        cmbSourcePic.ValueMember = "Value"
        cmbSourcePic.DataSource = _Pictures
        'SplitContainer2.Panel1.BackgroundImage an dieselbe DataSource binden
        Me.SplitContainer2.Panel1.DataBindings.Add("BackgroundImage", _Pictures, "Value")
        Reload()
        ApplyCurrentFilter()
    End Sub

    Private Sub Button_Click(ByVal sender As Object, ByVal e As EventArgs) _
          Handles btSave.Click, btReload.Click, btApplyFilter.Click, btRemoveFilter.Click
        Select Case True
            Case sender Is btSave
                Save()
            Case sender Is btReload
                Reload()
            Case sender Is btApplyFilter
                Me.DataGridView1.EndEdit()
                Me.FilterBindingSource.EndEdit()
                ApplyCurrentFilter()
            Case sender Is btRemoveFilter
                ExchangeDisposable( _
                    SplitContainer1.Panel2.BackgroundImage, _
                    New Bitmap(DirectCast(cmbSourcePic.SelectedValue, Image)))
        End Select
    End Sub

    Private Sub FilterTable_TableNewRow(ByVal sender As Object, ByVal e As DataTableNewRowEventArgs)
        'Filtermatrix neuer Datenstze initialisieren
        With DirectCast(e.Row, FilterRow)
            .Matrix = New SByte() {0, 0, 0, 0, 1, 0, 0, 0, 0}
        End With
    End Sub

    Private Sub Save()
        Me.DataGridView1.EndEdit()
        Me.FilterBindingSource.EndEdit()
        Me.FilterDataSet.WriteXml(_DataFile)
    End Sub

    Private Sub Reload()
        If Not File.Exists(_DataFile) Then
            MsgBox("Es sind noch keine Filterdaten abgespeichert worden")
            Return
        End If
        Me.FilterDataSet.Filter.BeginLoadData()
        Me.FilterDataSet.Filter.Clear()
        Me.FilterDataSet.ReadXml(_DataFile)
        Me.FilterDataSet.Filter.EndLoadData()
    End Sub

    Private Sub ApplyCurrentFilter()
        'ber jedes Pixel wird eine 3*3-Gewichtungs-Matrix gelegt, die darunter liegenden Werte
        '! gewichtet aufsummiert. Die Gesamtsumme wird durch den Divisor geteilt, um den
        '! Durchschnitts-Wert wieder herzustellen. 
        'Heben sich positive und negative Gewichtungen auf (Summe der Matrix = 0) so wird die
        '! Durchschnitts-Helligkeit des Bildes gewhrleistet durch Addition eines Offsets von 127.
        If FilterBindingSource.Position < 0 Then Return
        Dim Filter As FilterRow = GetRow(Of FilterRow)(Me.FilterBindingSource.Current)
        Dim Img As Image = DirectCast(cmbSourcePic.SelectedValue, Image)
        Dim Helper As New PixelHelper(Img, PixelFormat.Format24bppRgb, ckGray.Checked)
        Dim Divisor As Single = Filter.Divisor
        Dim Offset As Short = Filter.Offset
        Dim ReadPixels As Byte(,) = Helper.ReadPixels
        Dim rpX, rpY As Integer
        Dim WritePixels As Byte(,) = Helper.CreateWritePixels(PixelFormat.Format24bppRgb)
        Dim wpX, wpY As Integer
        Dim Matrix As SByte(,) = Me.UclMatrixGrid1.GetMatrix
        Dim mtrX, mtrY As Integer
        Dim YUbound As Integer = Img.Height - 1
        Dim XUbound As Integer = Img.Width * 3 - 1
        'Checkboxen steuern, welche Farbkanle filtern, oder ob Grauwert-Konvertierung anwenden
        Dim FilterColor As Boolean() = New Boolean() {ckRed.Checked, ckGreen.Checked, ckBlue.Checked}
        If ckGray.Checked Then FilterColor = New Boolean() {True, True, True}
        For IColor As Integer = 0 To 2
            If FilterColor(IColor) Then
                'die Hauptschleife lt die Rnder aus, da die Anwendung der Matrix auf Randpixel
                '! zu Array-berschreitungen fhrt (s. Start- und End-Werte von wpY, wpX).
                For wpY = 1 To YUbound - 1
                    'horizontal in 3-er Schritten iterieren, da die 3 FarbKomponenten der Pixel
                    '! hintereinander liegen
                    For wpX = IColor + 3 To XUbound - 3 Step 3
                        Dim WeightedSum As Double = 0
                        For mtrY = 0 To 2
                            'Vertikalversatz zum Auslesen der Nachbarpixel
                            rpY = wpY + mtrY - 1
                            For mtrX = 0 To 2
                                'Horizontalversatz in 3-er Schritten
                                rpX = wpX + (mtrX - 1) * 3
                                WeightedSum += ReadPixels(rpY, rpX) * Matrix(mtrY, mtrX)
                            Next
                        Next
                        WeightedSum = ClipBetween(0.0, WeightedSum / Divisor + Offset, 255.0)
                        WritePixels(wpY, wpX) = CByte(WeightedSum)
                    Next
                Next
                'oberer und unterer Rand, incl. Eck-Pixel
                For wpY = 0 To YUbound Step YUbound
                    For wpX = IColor To XUbound Step 3
                        Dim WeightedSum As Double = _
                           ApplyMatrixSave(ReadPixels, Matrix, wpY, wpX, YUbound, XUbound)
                        WritePixels(wpY, wpX) = CByte(ClipBetween(0.0, WeightedSum / Divisor + Offset, 255.0))
                    Next
                Next
                'linker und rechter Rand, Eck-Pixel auslassen
                For wpY = 1 To YUbound - 1
                    For wpX = IColor To XUbound Step XUbound - 2
                        Dim WeightedSum As Double = _
                           ApplyMatrixSave(ReadPixels, Matrix, wpY, wpX, YUbound, XUbound)
                        WritePixels(wpY, wpX) = CByte(ClipBetween(0.0, WeightedSum / Divisor + Offset, 255.0))
                    Next
                Next
            Else
                'statt filtern Farbkomponente nur kopieren
                For wpY = 0 To YUbound
                    For wpX = IColor To XUbound Step 3
                        WritePixels(wpY, wpX) = ReadPixels(wpY, wpX)
                    Next
                Next
            End If
        Next
        ExchangeDisposable(Me.SplitContainer1.Panel2.BackgroundImage, Helper.GetResultBitmap)
    End Sub


    Private Function ApplyMatrixSave( _
              ByVal ReadPixels As Byte(,), _
              ByVal Matrix As SByte(,), _
              ByVal wpY As Integer, _
              ByVal wpX As Integer, _
              ByVal YUbound As Integer, _
              ByVal XUbound As Integer) As Double
        'wendet die Matrix auf ein Pixel an, unter Absicherung gegen Arrayberschreitungen in
        '! alle 4 Richtungen
        Dim WeightedSum As Double = 0
        For mtrY As Integer = 0 To 2
            'ReadPixel-Y wird in den Y-Array-Grenzen geclippt
            Dim rpY As Integer = ClipBetween(0, wpY + mtrY - 1, YUbound)
            For mtrX As Integer = 0 To 2
                Dim rpX As Integer = wpX + (mtrX - 1) * 3
                'Bei Randberschreitung mu rpX um 3 Bytes versetzt werden, um dieselbe
                '! Farbkomponente des horizontalen Nachbarpixels auszulesen 
                If rpX < 0 Then
                    rpX += 3
                ElseIf rpX > XUbound Then
                    rpX -= 3
                End If
                WeightedSum += ReadPixels(rpY, rpX) * Matrix(mtrY, mtrX)
            Next
        Next
        Return WeightedSum
    End Function

    Private Sub FilterBindingSource_CurrentChanged(ByVal sender As Object, ByVal e As EventArgs) _
            Handles FilterBindingSource.CurrentChanged, cmbSourcePic.SelectedValueChanged, _
            ckRed.CheckedChanged, ckGreen.CheckedChanged, ckBlue.CheckedChanged, ckGray.CheckedChanged
        ApplyCurrentFilter()
    End Sub

End Class
